#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <neaacdec.h>
#include <sys/select.h>
#include <inttypes.h>
#include "aac.h"
#include "alsa.h"
#include "download.h"
#include "utils.h"
#include "audio_buffer.h"
#include "logger.h"
#include "mp4parser.h"

void *audio_thread(void *_d)
{
  audio_data *d = _d;
  char *full_url;
  dbuf b;
  dbuf b_mdat;
  NeAACDecHandle h;
  char *audio_config;
  int audio_config_size;
  unsigned long sample_rate;
  unsigned char channels;
  alsa_data ad;
  uint64_t audio_start;
  uint64_t next_in_tfrf;

  NeAACDecConfiguration aac_config;

  if (d->audio_priv == NULL || d->audio_url == NULL) return NULL;

  audio_config = malloc(strlen(d->audio_priv));
  if (audio_config == NULL) abort();
  audio_config_size = hex2raw(audio_config, d->audio_priv);

  h = NeAACDecOpen();

  aac_config.defObjectType = LC;
  aac_config.defSampleRate = 48000;
  aac_config.outputFormat = FAAD_FMT_FLOAT;
  aac_config.downMatrix = 1;
  aac_config.useOldADTSFormat = 0;
  if (NeAACDecSetConfiguration(h, &aac_config) == 0) abort();

  if (NeAACDecInit2(h, (unsigned char *)audio_config, audio_config_size,
                    &sample_rate, &channels) < 0) abort();

  logprintf("aac private %s sr %ld channels %d\n",
            d->audio_priv, sample_rate, channels);

  ad.samplerate = sample_rate;
  ad.channels = channels;
  alsa_init(&ad);

  b.size = b.maxsize = 0; b.data = NULL; b.error = 0;

  full_url = malloc(strlen(d->base_url) + strlen(d->audio_url) + 256);
  if (full_url == NULL) abort();

  audio_start = d->audio_start;
  while (!its_over) {
    char t[128];

    sprintf(full_url, "%s%s", d->base_url, d->audio_url);

    sprintf(t, "%"PRIu64, audio_start);
    replace(full_url, "{start time}", t);

    sprintf(t, "%"PRIu64, d->audio_bitrate);
    replace(full_url, "{bitrate}", t);

    do {
      download(&b, full_url);
      if (its_over) break;
      if (b.error == 1 || b.size == 0) {
        extern int over_read;
        struct timeval timeout;
        fd_set rset;
        FD_ZERO(&rset);
        FD_SET(over_read, &rset);
        timeout.tv_sec = 0;
        timeout.tv_usec = 100 * 1000;
        if (select(over_read+1, &rset, NULL, NULL, &timeout) == -1) abort();
        if (its_over) break;
      }
    } while (!its_over && (b.error == 1 || b.size == 0));
    logprintf("downloaded [len %d] %s\n", b.size, full_url);

    if (its_over) goto quit;

    next_in_tfrf = get_next_time(b.data, b.size);

    get_mdat(&b, &b_mdat);

    if (b_mdat.error == 0) while (b_mdat.size) {
      NeAACDecFrameInfo info;
      void *dat = NeAACDecDecode(h, &info,
                                (unsigned char *)b_mdat.data, b_mdat.size);
      if (dat == NULL || info.error) goto quit;
      logprintf("dat size %d bytesconsumed %ld samples %ld channels %d\n",
                b_mdat.size,
                info.bytesconsumed, info.samples, info.channels);
      if (info.samples == 0) /*free(dat)*/;
      else                   put_audio(dat, info.samples * sizeof(float));
      if (its_over) goto quit;
//fwrite(dat, info.samples * 2, 1, stdout); fflush(stdout);
      if (info.bytesconsumed > b_mdat.size) abort();
      b_mdat.data += info.bytesconsumed;
      b_mdat.size -= info.bytesconsumed;
    }

    //dump_audio(b.data, b.size);

    if (next_in_tfrf) audio_start = next_in_tfrf;
    else {
      printf("WARNING: audio start not found in moof tfrf, "
             "using inc %"PRIu64"\n", d->audio_inc);
      audio_start += d->audio_inc;
    }
  }

quit:
  alsa_close();

  NeAACDecClose(h);

  free(audio_config);
  free(full_url);
  free(b.data);

  return NULL;
}
