#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <curl/curl.h>
#include <sys/select.h>
#include <unistd.h>
#include "download.h"
#include "utils.h"
#include "logger.h"

#define CR do { \
  logprintf("%s:%d: CURL error\n", __FILE__, __LINE__); \
  abort(); \
  } while (0)

static size_t data_in(const void *ptr, size_t size, size_t nmemb, void *_b)
{
  dbuf *b = _b;
  size_t l = size * nmemb;
  if (b->size + l > b->maxsize) {
    b->maxsize += (l + 4095) & ~4095;
    b->data = realloc(b->data, b->maxsize); if (b->data == NULL) abort();
  }
  memcpy(b->data+b->size, ptr, l);
  b->size += l;
  return nmemb;
}

static void do_download(dbuf *b, char *url, int stoppable, int timeout_ms)
{
  CURL *c;
  CURLM *m;
  CURLMcode retcode;
  int running;
  fd_set rset, wset, eset;
  int fdmax;
  CURLMsg *msg;
  int queue_size;
  int max;
  long http;
  extern int over_read;

  b->size = 0;
  b->error = 0;
  c = curl_easy_init(); if (c == NULL) CR;
  if (curl_easy_setopt(c, CURLOPT_URL, url)) CR;
  if (curl_easy_setopt(c, CURLOPT_HEADER, 0)) CR;
  if (curl_easy_setopt(c, CURLOPT_FOLLOWLOCATION, 1)) CR;
  if (curl_easy_setopt(c, CURLOPT_WRITEFUNCTION, data_in)) CR;
  if (curl_easy_setopt(c, CURLOPT_FILE, b)) CR;
  if (timeout_ms > 0)
    if (curl_easy_setopt(c, CURLOPT_TIMEOUT_MS, timeout_ms)) CR;
  //if (curl_easy_perform(c)) b->error = 1;

  /* let's go multi to abort ASAP if TV channel changed by user
   * (signaled with a byte present in the pipe 'over_read')
   */
  m = curl_multi_init(); if (m == NULL) CR;
  if (curl_multi_add_handle(m, c) != CURLM_OK) CR;
  retcode = curl_multi_perform(m, &running);
  if (retcode != CURLM_CALL_MULTI_PERFORM && retcode != CURLM_OK) CR;
  if (running == 0) goto done;
  while (running) {
    FD_ZERO(&rset);
    FD_ZERO(&wset);
    FD_ZERO(&eset);
    if (curl_multi_fdset(m, &rset, &wset, &eset, &fdmax) != CURLM_OK) CR;
    if (stoppable) FD_SET(over_read, &rset);
    if (fdmax == -1) {
      struct timeval tval;
      tval.tv_sec = 0;
      tval.tv_usec = 10 * 1000;
      if (stoppable) {
        if (select(over_read+1, &rset, NULL, NULL, &tval) == -1) abort();
        if (FD_ISSET(over_read, &rset)) goto over;
      } else {
        if (select(0, NULL, NULL, NULL, &tval) == -1) abort();
      }
      retcode = curl_multi_perform(m, &running);
      if (retcode != CURLM_CALL_MULTI_PERFORM && retcode != CURLM_OK) CR;
      continue;
    }
    max = stoppable ? (fdmax > over_read ? fdmax : over_read) : fdmax;
    if (select(max + 1, &rset, &wset, &eset, NULL) == -1) abort();
    if (stoppable && FD_ISSET(over_read, &rset)) goto over;
    retcode = curl_multi_perform(m, &running);
    if (retcode != CURLM_CALL_MULTI_PERFORM && retcode != CURLM_OK) CR;
  }
done:
  msg = curl_multi_info_read(m, &queue_size); if (msg == NULL) abort();
  if (msg->msg != CURLMSG_DONE) abort();
  if (msg->easy_handle != c) abort();

  if (msg->data.result) b->error = 1;

  if (curl_easy_getinfo(msg->easy_handle,
                        CURLINFO_RESPONSE_CODE, &http) == CURLE_OK) {
    if (http >= 400) b->error = 1;
  }

  if (curl_multi_remove_handle(m, c) != CURLM_OK) CR;
  curl_easy_cleanup(c);
  if (curl_multi_cleanup(m) != CURLM_OK) CR;

  data_in("", 1, 1, b);
  b->size--;
  return;

over:
  if (curl_multi_remove_handle(m, c) != CURLM_OK) CR;
  curl_easy_cleanup(c);
  if (curl_multi_cleanup(m) != CURLM_OK) CR;

  b->size = 0;
}

void download(dbuf *b, char *url)
{
  do_download(b, url, 1, 0);
}

void download_unstoppable(dbuf *b, char *url)
{
  do_download(b, url, 0, 0);
}

void download_timed(dbuf *b, char *url, int timeout_ms)
{
  int try = 0;
  b->error = 0;
  b->size = 0;
  while (!its_over && (b->error == 1 || b->size == 0)) {
printf("try %d\n", try++);
    do_download(b, url, 1, timeout_ms);
  }
printf("over!\n");
}

void init_download(void)
{
  if (curl_global_init(CURL_GLOBAL_ALL)) CR;
}
