#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <pthread.h>
#include <sys/select.h>
#include <inttypes.h>
#include "download.h"
#include "utils.h"
#include "xml.h"
#include "h264.h"
#include "utils.h"
#include "aac.h"
#include "thread.h"
#include "video_buffer.h"
#include "audio_buffer.h"
#include "orange-channels.h"
#include "font.h"
#include "text_over.h"
#include "conf.h"
#include "logger.h"
#include "mp4parser.h"

static int volatile next_channel_requested;

static pthread_mutex_t x_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t x_cond = PTHREAD_COND_INITIALIZER;

int over_read;
static int over_write;

static volatile int current_channel = 0;

channels channels_list;

void next_channel(int delta)
{
  char z = 0;
  if (pthread_mutex_lock(&x_mutex)) abort();

  logprintf("delta %d\n", delta);
  /* not really necessary to optimize this loop, and my brain gets lazy */
  while (delta) {
    current_channel += delta>0 ? 1 : -1;
    if (current_channel < 0) current_channel = channels_list.size - 1;
    if (current_channel >= channels_list.size) current_channel = 0;
    delta -= delta>0 ? 1 : -1;
  }

  its_over = 1;
  if (write(over_write, &z, 1) != 1) abort();
  next_channel_requested = 1;
  /* lock/wait next_channel_requested == 0 */
  while (next_channel_requested != 0) {
    if (pthread_cond_wait(&x_cond, &x_mutex)) abort();
  }
  if (pthread_mutex_unlock(&x_mutex)) abort();
}

int set_channel(int channel)
{
  if (channel == current_channel) return 0;

  next_channel(channel-current_channel);

  return 1;
}

int get_current_channel(void)
{
  return current_channel;
}

void wakeup_x(void)
{
  if (pthread_mutex_lock(&x_mutex)) abort();
  next_channel_requested = 0;
  if (pthread_cond_broadcast(&x_cond)) abort();
  if (pthread_mutex_unlock(&x_mutex)) abort();
}

static void *video_decompress_thread(void *_vdata)
{
  video_data *vdata = _vdata;
  compressed_images_item im;

  while (!its_over) {
    im = get_compressed_images();
    if (its_over) { free(im.data); break; }
    dump_video(vdata, im.data, im.size);
    free(im.data);
  }

  return NULL;
}

static void usage(void)
{
  printf("usage: [options]\n");
  printf("options:\n");
  printf("    -v\n");
  printf("        be verbose\n");
  printf("    -s <service plan URL (eg. file:///your/servicePlan.json)>\n");
  printf("        use given URL for service plan instead of default\n");
  printf("        (%s)\n", SERVICE_PLAN);
  printf("    -t <timeout (ms)>\n");
  printf("        EXPERIMENTAL!\n");
  printf("        to improve zapping I use -t 100, that gives 100ms to\n");
  printf("        download the channel information after which if the\n");
  printf("        download has not completed, it is restarted.\n");
  printf("        When I change the channel I very often have to wait too\n");
  printf("        long for the new channel to come.\n");
  printf("        This timeout makes things better.\n");
  printf("        It's not perfect because downloading video or audio data\n");
  printf("        maybe also take too long.\n");
  printf("        Maybe a better mechanism is required...\n");
  printf("        (HTTP to serve audio/video is a clear sign of insanity.)\n");
  exit(0);
}

int main(int n, char **v)
{
  char *manifestURL = NULL;
  char *baseURL = NULL, *p;
  char *fullURL = NULL;
  dbuf b;
  xmldata x;
  uint64_t video_start;
  video_data vdata;
  pthread_t video_dec, audio_dec;
  void *tret;
  int pip[2];
  int valid_channel;
  int i;
  int first_10_channels = 0;
  extern char *service_plan;
  int timeout_ms = 0;
  uint64_t next_in_tfrf;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-v"))          { logger_on = 1; continue; }
    if (!strcmp(v[i], "-s") && i<n-1) { service_plan = v[++i]; continue; }
    if (!strcmp(v[i], "-t") && i<n-1) { timeout_ms = atoi(v[++i]); continue; }
    usage();
  }

  init_image_buffer();

  { extern void *xthread(void *); new_thread(xthread, NULL); }


  if (pipe(pip)) abort();
  over_read = pip[0];
  over_write = pip[1];

  do {
    char s[FILENAME_MAX+1];
    char *home = getenv("HOME");
    if (home == NULL) break;
    /* 20 is arbitrary, bigger than strlen("/.orangetv.conf") */
    if (strlen(home) > FILENAME_MAX-20) abort();
    sprintf(s, "%s/.orangetv.conf", home);
    init_conf(s);
  } while (0);

  init_font();

  oclear();
  coprintf(128, 255, 128, "loading channels...\n");

  b.size = b.maxsize = 0; b.data = NULL; b.error = 0;
  init_download();

  global_init_video();

  memset(&channels_list, 0, sizeof(channels_list));
  get_channels(&channels_list);
  if (channels_list.size == 0) { printf("no channels?\n"); return 1; }

  { extern volatile int x_can_go; x_can_go = 1; }
  { void force_x_redraw(void); force_x_redraw(); }

restart:
  logprintf("%s\n", channels_list.c[current_channel].name);
  oclear();
  if (first_10_channels != 10) {
    coprintf(128, 128, 255, "[press h for help]\n");
    first_10_channels++;
  }
  oprintf("%s\n", channels_list.c[current_channel].name);
  if (channels_list.c[current_channel].drm)
    coprintf(255, 128, 128, "[channel might be 'protected' by DRM]\n");
  lock_channels_list();
  /* let's print the current program */
  if (channels_list.c[current_channel].psize) {
    channel *cc = &channels_list.c[current_channel];
    int cp;
    time_t t = time(NULL);
    for (cp = 0; cp < cc->psize; cp++)
      if (t >= cc->p[cp].start && t <= cc->p[cp].end) break;
    /* special case: the guide starts in the future, show it anyway */
    if (t < cc->p[0].start) cp = 0;
    if (cp == cc->psize)
      coprintf(128, 128, 255, "[reload the guide: ctrl+g]\n");
    else
      if (cc->p[cp].title)
        coprintf(128, 255, 128,
                 "[%2.2dh%2.2d - %2.2dh%2.2d] %s\n",
                 cc->p[cp].hstart, cc->p[cp].mstart,
                 cc->p[cp].hend, cc->p[cp].mend,
                 cc->p[cp].title);
  }
  unlock_channels_list();
  oprintf("loading...\n");

  /* let's run even if no channel present (ugly code, I'm not proud) */
  valid_channel = 1;
  manifestURL = get_channel_url(&channels_list.c[current_channel]);
  if (manifestURL == NULL) valid_channel = 0;
  if (manifestURL != NULL) baseURL = strdup(manifestURL);
  else                     baseURL = NULL;
  if (manifestURL != NULL && baseURL == NULL) abort();
  if (baseURL != NULL) {
    p = strstr(baseURL, "/Manifest");
    if (p == NULL) { logprintf("no /Manifest in your URL\n"); exit(1); }
    else p[1] = 0;
  }

  b.size = 0;
  if (manifestURL != NULL) download_timed(&b, manifestURL, timeout_ms);
  free(manifestURL); manifestURL = NULL;

  if (b.size == 0) valid_channel = 0;

  memset(&x, 0, sizeof(x));
  if (parse_xml(&x, b.data, b.size)) valid_channel = 0;

  logprintf("video start time %"PRIu64"\n", x.video_start);

  set_start_time();
  if (valid_channel) {
    set_video_start_time((double)x.video_start / x.video_timescale * 1000.);
    set_audio_start_time((double)x.audio_start / x.audio_timescale * 1000.);
  }

  audio_data ad;
  ad.audio_start   = x.audio_start;
  ad.audio_inc     = x.audio_inc;
  ad.audio_bitrate = x.audio_bitrate;
  ad.base_url      = baseURL;
  ad.audio_url     = x.audio_url;
  ad.audio_priv    = x.audio_priv;
  audio_dec = new_thread(audio_thread, &ad);

  video_dec = new_thread(video_decompress_thread, &vdata);

  init_video_data(&vdata);

  if (valid_channel == 0) goto over;

  fullURL = malloc(strlen(baseURL) + strlen(x.video_url) + 128);
  if (fullURL == NULL) abort();

  dump_video_header(&vdata, x.video_priv);

  oup();
  { void force_x_redraw(void); force_x_redraw(); }

  odelay(1000);

  video_start = x.video_start;
  while (!its_over) {
    char t[128];

    sprintf(fullURL, "%s%s", baseURL, x.video_url);

    sprintf(t, "%"PRIu64, video_start);
    replace(fullURL, "{start time}", t);

    sprintf(t, "%"PRIu64, x.video_bitrate);
    replace(fullURL, "{bitrate}", t);

    do {
      download(&b, fullURL);
      if (its_over) break;
      if (b.error == 1 || b.size == 0) {
        extern int over_read;
        struct timeval timeout;
        fd_set rset;
        FD_ZERO(&rset);
        FD_SET(over_read, &rset);
        timeout.tv_sec = 0;
        timeout.tv_usec = 100 * 1000;
        if (select(over_read+1, &rset, NULL, NULL, &timeout) == -1) abort();
        if (its_over) break;
      }
    } while (!its_over && (b.error == 1 || b.size == 0));
    logprintf("downloaded [len %d] %s\n", b.size, fullURL);

    if (its_over) break;

    next_in_tfrf = get_next_time(b.data, b.size);

    put_compressed_images(b.data, b.size);

    if (next_in_tfrf) video_start = next_in_tfrf;
    else {
      printf("WARNING: video start not found in moof tfrf, "
             "using inc %"PRIu64"\n",
             x.video_inc);
      video_start += x.video_inc;
    }
  }

over:
  if (valid_channel == 0) {
    oup();
    coprintf(255, 128, 128, "no signal");
    { void force_x_redraw(void); force_x_redraw(); }
    fd_set rset;
    FD_ZERO(&rset);
    FD_SET(over_read, &rset);
    while (1) {
      if (select(over_read+1, &rset, NULL, NULL, NULL) == -1) abort();
      if (FD_ISSET(over_read, &rset)) break;
    }
  }

  its_over_buffer();
  its_over_buffer_audio();

  if (pthread_join(audio_dec, &tret)) abort();
  if (pthread_join(video_dec, &tret)) abort();

  reset_buffer();
  reset_audio_buffer();

  free(baseURL); baseURL = NULL;
  free(x.video_url);
  free(x.audio_url);
  free(x.video_priv);
  free(x.audio_priv);
  free(fullURL); fullURL = NULL;
  free_video_data(&vdata);

  { char dumb; if (read(over_read, &dumb, 1) != 1) abort(); }

#if 0
  char *url;
  do {
    current_channel += next_channel_requested;
    if (current_channel >= c.size) current_channel = 0;
    if (current_channel < 0) current_channel = c.size - 1;
    url = get_channel_url(&c.c[current_channel]);
  } while (c.c[current_channel].drm == 1 || url == NULL);
  free(url);
#endif

  its_over = 0;

  wakeup_x();

  goto restart;

  return 0;
}
